#!/opt/fsr-elevate/elevate/.env/bin/python
import os, json, sys
import yaml
import argparse

# Ensure the RabbitMQ cluster directory is in sys.path
# because this script is called from cyops-common, which may not include it by default
sys.path.append("/opt/cyops-rabbitmq/rabbitmq-cluster")
from utils import read_config_file
from message_constants import *
from consts import *
from common_utils import take_yes_no_input, validate_hostnames, open_hosts_and_ports_in_firewall, is_mtls_enabled, \
    print_dataset, print_error, get_fqdn, take_default_config_backup
from utils import refresh_mq_certs, get_mq_ca_cert, get_mq_tcp_port, get_mq_tls_port, get_mq_ca_key, \
    generate_mq_client_certs, get_file_contents, update_current_config_for_join_cluster, update_current_config
from mq_utils import replace_erlang_cookie, stop_rabbitmq_server, restart_rabbitmq_server, \
    stop_rabbitmq_app, perform_rabbitmq_list_policies, perform_rabbitmq_reset, perform_rabbitmq_create_policy, \
    perform_join_cluster, start_rabbitmq_app, update_cyops_user_password, get_rabbitmq_nodename, \
    perform_rabbitmq_forget_node, get_cyops_user_password, execute_status_cmd, execute_cluster_status_cmd, \
    delete_rabbitmq_policy, parse_definition_args, perform_rabbitmq_leave_cluster, perform_rabbitmq_list_vhost, \
    stop_rabbitmq_app_and_reset
from ssh import get_ssh_config, get_primary_config_using_ssh, allowlist_on_remote_server
from logger import get_mq_logger
logger = get_mq_logger(__name__)

hide_clustering_cmds = True

def add_subparser(subparsers, name, description, hidden_cmd):
    kwargs = {"description": description}
    if not hidden_cmd:
        kwargs["help"] = description
    return subparsers.add_parser(name, **kwargs)

def add_join_cluster_subparser(subparsers):
    parser_join_cluster = add_subparser(subparsers, "join-cluster", "Join a node to the cluster", hidden_cmd=hide_clustering_cmds)
    parser_join_cluster.set_defaults(operation='join-cluster')
    parser_join_cluster.add_argument('--config-file', action='store', type=str, dest='config_file_path', default=None,
                                     help='Conf file copied from primary server')
    parser_join_cluster.add_argument('--primary-node', action='store', type=str, dest='primary_node_hostname',
                                     default=None,
                                     help='Primary node DNS name')
    parser_join_cluster.add_argument('--primary-node-ssh-key', action='store', type=str,
                                     dest='primary_node_ssh_key_path',
                                     default=None,
                                     help='SSH private key of primary node')

def add_leave_cluster_subparser(subparsers):
    leave_parser = add_subparser(subparsers, "leave-cluster", "Remove a node from the cluster", hidden_cmd=hide_clustering_cmds)
    leave_parser.set_defaults(operation='leave-cluster')
    leave_parser.add_argument('--no-interaction', action='store_false', dest='interactive', default=True,
                              help="Don't ask for any confirmation")
    # leave_parser.add_argument("--primary", required=True, help="Primary node of the cluster.")
    # leave_parser.add_argument("--node", required=True, help="Node to be removed from cluster.")

def add_forget_node_subparser(subparsers):
    parser_forgetnode = add_subparser(subparsers, "forget-node", "Remove a node from the HA cluster", hidden_cmd=hide_clustering_cmds)
    parser_forgetnode.set_defaults(operation="forget-node")
    parser_forgetnode.add_argument('--nodename', action='store', type=str, required=True)
    parser_forgetnode.add_argument('--no-interaction', action='store_false', dest='interactive', default=True,
                                     help="Don't ask for any confirmation")


def add_cluster_status_subparser(subparsers):
    cluster_status_parser = add_subparser(subparsers, "cluster-status", "Get the status of the cluster", hidden_cmd=True)
    cluster_status_parser.set_defaults(operation='cluster-status')


def add_status_subparser(subparsers):
    status_parser = add_subparser(subparsers, "status", "Get the status of the node.", hidden_cmd=True)
    status_parser.set_defaults(operation='status')


def add_list_cluster_nodes_subparser(subparsers):
    nodes_parser = add_subparser(subparsers, "list-nodes", "List all nodes in the cluster", hidden_cmd=hide_clustering_cmds)
    nodes_parser.set_defaults(operation='list-nodes')


def add_allow_ports_subparser(subparsers):
    port_parser = add_subparser(subparsers, "allowlist", "Enables ports of nodes for RabbitMQ cluster communication", hidden_cmd=hide_clustering_cmds)
    port_parser.set_defaults(operation='allowlist')
    port_parser.add_argument("--nodes", nargs="+", action='store', type=str, required=True,
                             help="Comma-separated list of hostnames of the cluster nodes")


def add_export_config_subparser(subparsers):
    config_parser = add_subparser(subparsers, "export-conf", "Export configuration details to form cluster", hidden_cmd=hide_clustering_cmds)
    config_parser.set_defaults(operation='export-conf')


def add_create_policy_subparser(subparsers):
    policy_parser = add_subparser(subparsers, "create-policy", "Create a new RabbitMQ policy", hidden_cmd=True)
    policy_parser.set_defaults(operation='create-policy')
    policy_parser.add_argument("--name", required=True, help="The name of the policy")
    policy_parser.add_argument("--pattern", required=True, help="The regex pattern for the policy")
    policy_parser.add_argument("--priority", default=0, help="The priority for the policy.")
    policy_parser.add_argument("--vhost", default='/', help="The virtual host (default: '/')")
    policy_parser.add_argument(
        "--apply-to",
        default='all',
        dest="apply_to",
        choices=['all', 'queues', 'exchanges', 'classic-queues', 'quorum-queues', 'streams'],
        help="The type of RabbitMQ objects the policy applies to (default: all)"
    )
    policy_parser.add_argument(
        "--definition",
        nargs='+',
        default=[],
        required=True,
        help="Policy definitions as key=value pairs (e.g., ha-mode=all ha-sync-mode=automatic)"
    )


def add_delete_policy_subparser(subparsers):
    policy_parser = add_subparser(subparsers, "delete-policy", "Deletes RabbitMQ policy", hidden_cmd=True)
    policy_parser.set_defaults(operation='delete-policy')
    policy_parser.add_argument("--name", help="The name of the policy")
    policy_parser.add_argument("--vhost", default='/', help="The virtual host (default: '/')")


def add_list_policy_subparser(subparsers):
    policy_parser = add_subparser(subparsers, "list-policies", "Retrieves list of RabbitMQ policies.", hidden_cmd=True)
    policy_parser.set_defaults(operation='list-policies')
    policy_parser.add_argument("--vhost", help="The virtual host (default: '/')")


def define_parser_cluster(subparsers, is_sme=True, parser=None):
    if is_sme:
        add_join_cluster_subparser(subparsers)
        add_leave_cluster_subparser(subparsers)
        add_forget_node_subparser(subparsers)
        add_export_config_subparser(subparsers)
        add_allow_ports_subparser(subparsers)
        add_list_cluster_nodes_subparser(subparsers)
    add_cluster_status_subparser(subparsers)
    add_status_subparser(subparsers)
    add_create_policy_subparser(subparsers)
    add_delete_policy_subparser(subparsers)
    add_list_policy_subparser(subparsers)
    if hide_clustering_cmds:
        options_to_hide = ["join-cluster", "list-nodes", "cluster-status", "status", "leave-cluster", "export-conf",
                           "allowlist", "forget-node", "create-policy", "delete-policy", "list-policies"]
        hide_command_options(parser, subparsers, options_to_hide)
    else:
        options_to_hide = ["cluster-status", "status", "create-policy", "delete-policy", "list-policies"]
        hide_command_options(parser, subparsers, options_to_hide)

        visible_options = "{join-cluster, list-nodes, leave-cluster, export-conf, allowlist, forget-node}"
        subparsers.metavar = visible_options


def define_parser(parser):
    subparsers = parser.add_subparsers(dest="operation")
    global hide_clustering_cmds
    # Do not hide clustering command options for external SME when called as, i.e., csadm ha --help
    hide_clustering_cmds = False
    define_parser_cluster(subparsers, parser=parser)
    for name, subp in subparsers.choices.items():
        subp._positionals.title = 'subcommand options are'
        subp._optionals.title = 'subcommand options are'


def hide_command_options(parser, subparsers, options_to_hide):
    try:
        def silent_error(msg):
            """Override argparse error to suppress default error message."""
            raise SystemExit(2)

        # Backup original argparse functions
        original_error = parser.error
        original_print_help = parser.print_help

        # Override argparse error and help functions to prevent duplicate messages
        parser.error = silent_error
        parser.print_help = lambda *args, **kwargs: None

        # Parse arguments
        args = parser.parse_args()
    except SystemExit:
        # Remove hidden commands from subparsers
        subparser_choices = getattr(subparsers, "choices", None)
        if subparser_choices:
            for option in options_to_hide:
                subparser_choices.pop(option, None)
    finally:
        # Restore original argparse functions
        parser.error = original_error
        parser.print_help = original_print_help


def join_cluster(config):
    logger.info("Performing join cluster")
    primary_node = config[MQ_SECTION][MQ_HOST]
    logger.debug(f"Joining cluster with {primary_node}")
    local_config = read_config_file()
    local_config[MQ_SECTION][MQ_PWD] = config[MQ_SECTION][MQ_PWD]
    local_config[MQ_SECTION][MQ_USER] = config[MQ_SECTION][MQ_USER]
    local_config[MQ_SECTION][MQ_VHOST] = config[MQ_SECTION][MQ_VHOST]
    take_default_config_backup()
    update_current_config_for_join_cluster(config, local_config)
    # At this point since we have the ca cert and key of primary
    # need to refresh the MQ certs of secondary
    refresh_mq_certs()

    # Sometimes RabbitMQ does not restart post erlang cookie is modified. Ensure no data is present.
    # As it is just precautionary operation pass ignore_error as True
    stop_rabbitmq_app_and_reset(ignore_error=True)

    # Replace the erlang cookie and start app
    stop_rabbitmq_server()
    replace_erlang_cookie(config[MQ_SECTION].get(ERLANG_COOKIE))
    restart_rabbitmq_server()

    # Perform mq join cluster
    stop_rabbitmq_app_and_reset()
    perform_join_cluster(config[MQ_SECTION][NODENAME])
    start_rabbitmq_app()
    cyops_password = config[MQ_SECTION][CYOPS_PASS]
    update_cyops_user_password(cyops_password)
    local_config[MQ_SECTION][ERLANG_COOKIE] = config[MQ_SECTION][ERLANG_COOKIE]
    local_config[MQ_SECTION][NODENAME] = config[MQ_SECTION][NODENAME]
    local_config[MQ_SECTION][CYOPS_PASS] = config[MQ_SECTION][CYOPS_PASS]

    # Node joined successfully, node must point to local MQ
    local_config[MQ_SECTION][MQ_HOST] = 'localhost'
    if MQ_HOST_PRIMARY in config[MQ_SECTION]:
        local_config[MQ_SECTION][MQ_HOST_PRIMARY] = config[MQ_SECTION][MQ_HOST_PRIMARY]
    local_config[MQ_SECTION][MQ_SSL_ENABLED] = False

    # Update current mq config
    update_current_config(local_config)

    # Update mq_password in default config
    default_config = read_config_file(DEFAULT_CYOPS_CONN_FILE)
    default_config[MQ_SECTION][MQ_PWD] = config[MQ_SECTION][MQ_PWD]
    update_current_config(default_config, file_path=DEFAULT_CYOPS_CONN_FILE)

class RabbitMQCluster:
    @staticmethod
    def leave_cluster(interactive=True) -> None:
        logger.info("Performing leave cluster")
        if interactive:
            input_ = take_yes_no_input(validation_msg="Are you sure you want to leave cluster?")
            if input_.lower() != 'yes':
                return
        perform_rabbitmq_leave_cluster()

    # @classmethod
    @staticmethod
    def list_cluster_nodes():
        raw_output = execute_cluster_status_cmd()
        cluster_status = json.loads(raw_output)
        disk_nodes = cluster_status.get('disk_nodes', [])
        running_nodes = set(cluster_status.get('running_nodes', []))
        node_info = [
            {
                "Node Name": node,
                "Status": "Running" if node in running_nodes else "Unknown",
                "RabbitMQ Version": cluster_status["versions"].get(node, {}).get("rabbitmq_version", "-"),
                "Maintenance Status": cluster_status["maintenance_status"].get(node, "Unknown"),
            }
            for node in disk_nodes
        ]
        return node_info

    @staticmethod
    def allow_list(hosts):
        mq_tls_port = get_mq_tls_port() or 5671
        mq_ports = [mq_tls_port, 4369, 25672, "35672-35682"]
        if isinstance(hosts, str):
            hosts = [host.strip() for host in hosts.split(',')]
        elif isinstance(hosts, (list, tuple)):
            processed_hosts = []
            for host_entry in hosts:
                split_hosts = str(host_entry).split(',')  # Ensure each entry is a string before splitting
                processed_hosts.extend([host.strip() for host in split_hosts])  # Trim spaces
            hosts = processed_hosts
        list_of_hosts = list(set(hosts))
        validate_hostnames(list_of_hosts)
        open_hosts_and_ports_in_firewall(list_of_hosts, mq_ports)
        logger.info("Successfully registered")\

    @staticmethod
    def forget_node(node_name, interactive=True):
        if interactive:
            input_ = take_yes_no_input(validation_msg=MSG_WARN_NODE_REMOVAL)
            if input_.lower() != 'yes':
                return
        stop_rabbitmq_app(node_name, ignore_error=True)
        perform_rabbitmq_forget_node(node_name, ignore_error=True)
        # Reset remote node
        perform_rabbitmq_reset(node_name, ignore_error=True)

    @staticmethod
    def exportconf(conf_file_path=None, create_conf_file=True):
        if conf_file_path is None:
            conf_file_path = '/home/csadmin/mq_ha.conf'

        fqdn_name = get_fqdn()
        db_config = read_config_file()

        # MQ Section
        db_config[MQ_SECTION][MQ_TCP_PORT] = get_mq_tcp_port()
        db_config[MQ_SECTION][MQ_TLS_PORT] = get_mq_tls_port()
        db_config[MQ_SECTION][MQ_CA_CERT] = get_mq_ca_cert()

        # CA key is needed so that we can use this key across cluster to generate server certs for MQ
        db_config[MQ_SECTION][MQ_CA_KEY] = get_mq_ca_key()

        # client-certs needs to be generated based on whether mTLS is enabled on the primary node.
        db_config[MQ_SECTION][MQ_MTLS_ENABLED] = False
        db_config[MQ_SECTION][ERLANG_COOKIE] = get_file_contents(ERLANG_COOKIE_PATH)[0]
        db_config[MQ_SECTION][NODENAME] = get_rabbitmq_nodename()
        db_config[MQ_SECTION][CYOPS_PASS] = get_cyops_user_password()
        if is_mtls_enabled():
            # Since mTLS is enabled, client certs are required for non localhost traffic
            db_config[MQ_SECTION][MQ_MTLS_ENABLED] = True
            client_certs_path = generate_mq_client_certs()
            if not client_certs_path:
                # Exit here since cluster formation without client-certs is not advisable since secondary's
                # celery and postman services will fail to connect to primary node
                logger.info(MSG_FAILED_TO_GENERATE_CLIENT_CERTS_FOR_MQ)
                sys.exit(1)
            if os.path.exists(client_certs_path + "/client-cert.pem"):
                with open(client_certs_path + "/client-cert.pem") as client_cert_file:
                    db_config[MQ_SECTION][MQ_CLIENT_CERT] = client_cert_file.read()
            else:
                logger.info(MSG_FAILED_TO_GENERATE_CLIENT_CERTS_FOR_MQ)
                sys.exit(1)
            if os.path.exists(client_certs_path + "/client-key.pem"):
                with open(client_certs_path + "/client-key.pem") as client_key_file:
                    db_config[MQ_SECTION][MQ_CLIENT_KEY] = client_key_file.read()
            else:
                logger.info(MSG_FAILED_TO_GENERATE_CLIENT_CERTS_FOR_MQ)
                sys.exit(1)

        # During export-conf, make the MQ_SSL_ENABLED flag as True
        db_config[MQ_SECTION][MQ_SSL_ENABLED] = True
        db_config[MQ_SECTION][MQ_HOST] = fqdn_name
        db_config[MQ_SECTION][MQ_HOST_PRIMARY] = fqdn_name
        mq_config = {MQ_SECTION: db_config[MQ_SECTION]}
        if create_conf_file:
            with open(conf_file_path, 'w') as ha_conf_file:
                yaml.dump(mq_config, ha_conf_file, default_flow_style=False)
            print("Configuration details written to {0}".format(conf_file_path))
        return mq_config

    @staticmethod
    def check_status():
        raw_output = execute_status_cmd()
        node_status = json.loads(raw_output)
        table_data = [
            ["OS", node_status.get("os", "N/A")],
            ["RabbitMQ Version", node_status.get("rabbitmq_version", "N/A")],
            ["Erlang Version", node_status.get("erlang_version", "N/A")],
            ["Uptime (s)", node_status.get("uptime", "N/A")],
            ["Under Maintenance", "No" if not node_status.get("is_under_maintenance", False) else "Yes"],
            ["Processes Used",
             f'{node_status.get("processes", {}).get("used", "N/A")} / '
             f'{node_status.get("processes", {}).get("limit", "N/A")}'],
            ["Run Queue", node_status.get("run_queue", "N/A")],
            ["Disk Free",
             f'~{node_status.get("disk_free", 0) / (1024 ** 3):.1f} GB '
             f'(Limit: {node_status.get("disk_free_limit", 0)} MB)'],
            ["File Descriptors",
             f'{node_status.get("file_descriptors", {}).get("total_used", "N/A")} / '
             f'{node_status.get("file_descriptors", {}).get("total_limit", "N/A")}'],
            ["Sockets Used",
             f'{node_status.get("file_descriptors", {}).get("sockets_used", "N/A")} / '
             f'{node_status.get("file_descriptors", {}).get("sockets_limit", "N/A")}'],
            ["Virtual Hosts", node_status.get("totals", {}).get("virtual_host_count", "N/A")],
            ["Total Queues", node_status.get("totals", {}).get("queue_count", "N/A")],
            ["Total Connections", node_status.get("totals", {}).get("connection_count", "N/A")],
            ["Listeners",
             "\n".join(
                 [f'{i.get("port", "N/A")} ({i.get("purpose", "N/A")})' for i in node_status.get("listeners", [])])],
            ["Config File", "\n".join(node_status.get("config_files", ["N/A"]))],
            ["Log Files", "\n".join(node_status.get("log_files", ["N/A"]))],
            ["Plugins", "\n ".join(node_status.get("active_plugins", ["N/A"]))]
        ]
        return table_data

    @staticmethod
    def check_cluster_status():
        raw_output = execute_cluster_status_cmd()
        cluster_status = json.loads(raw_output)
        disk_nodes = cluster_status.get('disk_nodes', [])
        running_nodes = set(cluster_status.get('running_nodes', []))
        cluster_info = [
            {
                "Node Name": node,
                "Status": "Running" if node in running_nodes else "Unknown",
                "RabbitMQ Version": cluster_status.get("versions", {}).get(node, {}).get("rabbitmq_version", "-"),
                "Erlang Version": cluster_status.get("versions", {}).get(node, {}).get("erlang_version", "-"),
                "CPU Cores": cluster_status.get("cpu_cores", {}).get(node, 0),
                "Maintenance Status": cluster_status.get("maintenance_status", {}).get(node, "unknown"),
                "Ports": "\n".join(
                    f'{listener.get("port", "-")} ({listener.get("purpose", "-")})'
                    for listener in cluster_status.get("listeners", {}).get(node, [])
                ) or "-"
            }
            for node in disk_nodes
        ]
        return cluster_info

    @staticmethod
    def create_policy(name, pattern, vhost='/', apply_to='all', priority=0, **definition):
        result = perform_rabbitmq_create_policy(name, pattern, vhost=vhost, priority=priority, apply_to=apply_to,
                                               **definition)
        result = json.loads(result)
        return result

    @staticmethod
    def delete_policy(name, vhost='/'):
        result = delete_rabbitmq_policy(name, vhost)
        return result

    @staticmethod
    def list_policies(vhosts=None):
        if not vhosts:
            # get all vhosts
            vhosts = perform_rabbitmq_list_vhost()
        elif isinstance(vhosts, str):
            vhosts = [{"name": vhosts}]
        policy_list = []
        for vhost in vhosts:
            policies = perform_rabbitmq_list_policies(vhost.get('name'))
            policy_list.extend(policies)
        return policy_list


def start(args):
    cluster = RabbitMQCluster()
    if args.operation == "join-cluster":
        config = None
        if args.config_file_path:
            config = read_config_file(args.config_file_path)
        elif args.primary_node_hostname:
            primary_node = args.primary_node_hostname
            primary_node_ssh_key = args.primary_node_ssh_key_path
            ssh_details = get_ssh_config(config, primary_node, primary_node_ssh_key)
            if ssh_details is not None:
                primary_config = get_primary_config_using_ssh(ssh_details)
                allowlist_on_remote_server(ssh_details, False)
                config = primary_config
        else:
            print_error(mconsts_9)
            sys.exit(1)
        join_cluster(config)
        logger.info("Joined cluster successfully")
        return
    elif args.operation == "leave-cluster":
        return cluster.leave_cluster(args.interactive)
    elif args.operation == "cluster-status":
        result = cluster.check_cluster_status()
        print_dataset(result)
    elif args.operation == 'status':
        result = cluster.check_status()
        print_dataset(result)
    elif args.operation == "list-nodes":
        result = cluster.list_cluster_nodes()
        print_dataset(result)
    elif args.operation == 'export-conf':
        cluster.exportconf()
    elif args.operation == 'forget-node':
        cluster.forget_node(args.nodename, args.interactive)
    elif args.operation == 'allowlist':
        cluster.allow_list(args.nodes)
    elif args.operation == 'create-policy':
        definitions = parse_definition_args(args.definition)
        result = cluster.create_policy(args.name, args.pattern, vhost=args.vhost, priority=args.priority,
                                       apply_to=args.apply_to, **definitions)
        print(result)
    elif args.operation == 'delete-policy':
        result = cluster.delete_policy(args.name, args.vhost)
        print(result)
    elif args.operation == 'list-policies':
        result = cluster.list_policies(args.vhost)
        if result:
            print_dataset(result)
        else:
            print("No policy found.")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog="csadm ha", description="RabbitMQ Cluster Management Script")
    define_parser(parser)
    args = parser.parse_args()
    start(args)
